module GPUFunctions
contains
!!!!!!!!!!!!!!!!!!!!!!! SUBROUTINE THAT DOES THE JOB: See Caprioli (2012) and refs therein
!!!!!!!! Especially Amato-Blasi, 2005; Caprioli et al. 2009;
	 
     
     
     



                                            !inputs arrays, then result output arrays, then result output matrix then values for writing during DSA
attributes(global) subroutine launchToGPU(ageArray, RshArray,VshArray, N0Array, B0Array, T0, Pcr_guess, Pcr_max, xi_inj, delta, X0, &
                                            PcrOutputArray, T2OutputArray, B2OutputArray, FescOutputArray, RtotOutputArray, &
                                            p_outMatrix, f0_outMatrix, fesc_outMatrix)
    implicit none
    real*8 :: M0, X0Value !input values
    real*8, value :: T0, Pcr_guess, Pcr_max, xi_inj, delta, X0
    real*8, dimension(:) :: ageArray, RshArray,VshArray, N0Array, B0Array !input arrays (timestep) to pass as values
    real*8, dimension(:) :: PcrOutputArray, T2OutputArray, B2OutputArray, FescOutputArray, RtotOutputArray !result output arrays (timestep) for final output.dat 
    real*8, dimension(:,:) :: p_outMatrix, f0_outMatrix, fesc_outMatrix !result output matrix (timestep, N) for spectrum.dat
    integer :: n, i, j

    n = blockIdx%x
    i = threadIdx%x
    M0=VshArray(n)*dsqrt(1.67d-24/5.d0*3.d0/1.38d-16/T0)
    X0Value = X0*RshArray(n)
    
    if(i == 1)then !only launching on first thread for first version.
        call DSA_X0(M0,B0Array(n),T0,n0Array(n),xi_inj,X0Value,Pcr_guess,Pcr_max, PcrOutputArray,T2OutputArray,B2OutputArray,FescOutputArray,RtotOutputArray,p_outMatrix,f0_outMatrix,fesc_outMatrix, delta)
    endif
 end subroutine launchToGPU
     
                                    !1 ,2 ,3 ,4 , 5 ,6 ,  7      , 8     ,   9           ,10           , 11          ,   12          ,      13       ,  14 ,  15  ,   16   ,   17
attributes(device) subroutine DSA_X0(M0,B0,T0,n0,csi,X0,Pcr_guess,Pcr_max, PcrOutputArray,T2OutputArray,B2OutputArray,FescOutputArray,RtotOutputArray,p_outMatrix,f0_outMatrix,fesc_outMatrix, delta)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!	N,NN => Parameters which set the NUMBER OF POINTS in the momentum and space grid.	  !!!!
!!!!			N = NN = 40 is recommended, since larger values are often indistinguishable!!!!
!!!!				while, for instance, N = 30 leads to an error of about 20 %
!!!!			In order to change the grid size, REPLACE N=Nnew, NN=NNnew in the
!!!!			subroutine DSA_X0 and in functions funcexp, funcp1,funcxi.
!!!!    HINT:-- Two ERRORS warn if the solution is not in the given range of RsRt or of Pmax:
!!!!            we may need to readjust this                                                     !!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!	INPUT PARAMETERS:
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!	M0 => Sonic Mach number at upstream infinity IN THE SHOCK REFERENCE FRAME
!!!!	B0 => Magnetic field at upstream infinty (in Gauss)
!!!!    T0 => Temperature at upstream infinity (in K)
!!!!	n0 => proton density at upstream infinity (in particles/cm^3)
!!!!	csi => injection parameter, it regulates CR efficiency : TRY 3.7 (suggested values between 3 and 4)
!!!!           SMALLER values mean MORE EFFICIENT INJECTION
!!!!	X0 => Free-escape boundary (in pc) = Diffusion legnth of particles at Pmax
!!!!    Pcr_guess=Initial guess for the CR pressure (in units of rho u0^2)
!!!!		 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!	OUTPUT PARAMETERS:
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!	Pcr => Pressure in CRs at the shock (normalized to bulk pressure rho u0^2)
!!!!	T2 => Downstream temperature
!!!!	B2 => Downstream magnetic field	
!!!!	Fesc => Energy flux in escaping particles, normalized to 0.5\mp*N0*u0^3
!!!!    Rtot => Total compression ratio
!!!!	p_out => array, CR momenta (in units of mc)
!!!!	f0_out => array, CR distribution at the shock (4pi p^4 f(p)/n0)
!!!!    fesc_out => array, escaping CR distribution (phi_esc(p)/n0/u0)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  
  implicit none

  integer :: N,i,j,k,Nprove,NN,verb, timestep
  PARAMETER (N=50,NN=50)
  real*8 :: x(N),p(N),lnp(N),Up(N),U(N),tau(N),Upold(N)
  real*8 :: fi(N)
  real*8 :: xvec(NN),yvec(NN)
  real*8 :: xvec0(NN),yvec0(NN)
  real*8 :: xvec1(NN),yvec1(NN)
  real*8 :: mc,mp,me,c,kB,Fth,zeta
  real*8 :: Rsub,Rtot,Ssub,Stot,gammag,gammaCR,gammaB,B2,T2,Pinj,eta
  real*8 :: va0,Ma0,DB0,N0,qe,yr
  real*8 :: pi,Pstar,u0,par1,par2,csi0n
  real*8 :: rest,arest,restold,rdiff,rd0
  real*8 :: Y1,ts,chi,dB(N),rsrt,rsrt1,rsrt2
  real*8 :: rsrtold, tol, Diff, res, dummy
  real*8 :: lamg,lamCR,lamB,aa,bb,cc,dd,xR,xRnew,err
  real*8 :: Xmin,Xmax,Xstep,Ymin,Ymax,Ystep
  real*8 :: pp,rest1,rest2
  real*8 :: exppsi,Psi(N,N),Psip(N),f0(N)
  real*8 :: dvec(N,N),dvecold(N,N)
  real*8 :: csi_c(N),csic0,PB(N),phi(N),V(N)
  real*8 :: norm,pth,pMB(N),fMB(N)
  real*8 :: tmp(N),tmp2(N),fxp(N,N),fxpold(N,N),DU(N),Wxp(N,N)
  real*8 :: f0old(N),Bxp(N,N),Kx(N,N),Kxold(N,N),csic0_err
  real*8 :: Ftld(N), Ftldold(N),Fesc
  real*8 :: p_out(N),f0_out(N),fesc_out(N),Pcr
  real*8 :: funcexp,funcp0,funcp1,funcxi,divdif
  real*8 :: V2 ! hack


  !variables for GPU version, inputs as values, outputs as arrays 
  real*8, value :: T0, Pcr_guess, Pcr_max, csi, delta, X0, M0, B0 !input values
  real*8, device :: PcrOutputArray(:), T2OutputArray(:), B2OutputArray(:), FescOutputArray(:), RtotOutputArray(:) !result output arrays (timestep) for final output.dat 
  real*8, device :: p_outMatrix(:,:), f0_outMatrix(:,:), fesc_outMatrix(:,:) !result output matrix (timestep, N) for spectrum.dat







  integer :: jf,jn,errorNumber

  PARAMETER (pi=3.1415d0)
  PARAMETER (tol=1.d-4, par1=1.d-3,par2=6.d3) !#CONTROL par1 defines Xmin and par2 the maximum P in the grids!!! 			
  PARAMETER (gammag=5.d0/3.d0)
  PARAMETER (gammaCR=4.d0/3.d0) 
  PARAMETER (gammaB=3.d0/2.d0)   				      
  PARAMETER (mp=1.67d-24,me=9.1d-28,c=3.d10,kB=1.38d-16,qe=4.8d-10)

  COMMON/VEC/xvec,yvec
  COMMON/VEC0/xvec0,yvec0
  COMMON/VEC1/xvec1,yvec1

  !external funcexp,funcp0,funcp1,funcxi !is this still needed since this is all in the same module file?
  
  timestep = blockIdx%x

  verb=0 !=1 turns verbose output on

  mc=mp*c  						
  yr=60.d0*60.d0*24.d0*365.25d0	

  u0=M0*dsqrt(gammag*kB*T0/mp)

  va0=B0/dsqrt(4.d0*pi*N0*mp)
  Ma0=u0/va0  
  ! write(*,*) 'Ma0 = ', Ma0          
  DB0=mp*c**3.d0/3.d0/qe/B0  	  	  
  zeta=1.d0-max(1.d-6,dexp(-3.d5/u0)) ! damping paramer: kicks in at ~300km/s;    zeta=0, no damping

  rsrt=1.d0-Pcr_guess
  rsrt1=1.d0-Pcr_max !!!!#CONTROL (1-rsrt1)=50%  is the maximum efficiency allowed
  rsrt2=1.d0

  arest=1.d0
  rest=1.d0
  rest1=0.d0
  rest2=0.d0

  rsrtold=1.d1
  do while(arest.ge.5.d-2) !CONTROL paramter that ensures convergence 
     !PB(1)=2.d0*(1.d0-rsrt**(5.d0/4.d0))**2.d0/25.d0/rsrt**1.5d0*(1.d0-zeta)*delta ! Check 2012 paper
     !PB(1)=u0/2.d0/c*(1.d0-rsrt)/(gammaCR-1.d0)*delta
     PB(1)=3.d0/2.d0*(1.d0-rsrt)/Ma0*delta
     dummy=1.d0+1.d0/gammag/M0/M0-rsrt-1.d0/rsrt**gammag/gammag/M0/M0-PB(1) ! PCR without any assumptions
     !PB(1)=(1.d0-rsrt)/11.d0
     !PB(1)=delta*(1.d0-rsrt)*3.d0/(1.d0+2.d0*4.d0**2.d0) ! decompress downstream B field assuming r=4
     chi=(1.d0-rsrt-PB(1))/PB(1) ! #PHYSICS: gas heating is neglected (good if adiabatic EOS) Pcr1/PB1
     Fth=zeta*2.d0*gammag*(gammag-1.d0)/(2.d0*gammag+1.d0)  &!&
         *M0*M0/Ma0*(1.d0-rsrt**(gammag+5.d-1)) ! turbulent/non-adiabatic heating
     tau(1)=rsrt**(-gammag)/gammag/M0/M0*(1.d0+Fth)
     
     lamg=gammag/(gammag-1.d0)
     lamCR=gammaCR/(gammaCR-1.d0)  
     lamB=gammaB/(gammaB-1.d0)
     aa=4.d0/3.d0*rsrt**2.d0*PB(1)*(lamB-lamg)
     bb=-1.d0-2.d0*lamg/gammag/M0/M0
     cc=2.d0*lamg+2.d0*lamg/gammag/M0/M0+2.d0*dummy*(lamCR-lamg)+2.d0*PB(1)/3.d0*(lamB-lamg)
     dd=1.d0-2.d0*lamg
     
     xR=4.d0
     do
     	xRnew=xR-((aa*xR**3.d0+bb*xR**2.d0+cc*xR+dd)/(3.d0*aa*xR**2.d0+2.d0*bb*xR+cc))
     	err=1.d2*abs((xRnew-xR)/xR)
     	xR=xRnew
     	if (err < 1.d-6) exit
     enddo
     
     Rtot=xR  
     Rsub=Rtot*rsrt

!!!!	Calculation of T2 and Pinj
	 ! dummy=1.d0+1.d0/gammag/M0/M0-1.d0/Rtot-(1.d0-rsrt)*(1.d0+delta) !Pth,2
	 ! Pinj=csi*dsqrt(2.d0*dummy/Rtot)*mp*u0/mc
	 
	 Pinj=csi*mp*u0/Rsub/mc !hack RD: Pinj=csi_inj*mp*u1
     eta=(4.d0/3.d0)*(1.d0/dsqrt(pi))*(Rsub-1.d0)*(csi**3.d0)*dexp(-csi*csi)
	 T2=(mc*Pinj/csi)**2.d0/(2.d0*kB*mp)
	 
!!!!	Grid in log(x)
     Xmin=dlog(par1*3.d0*DB0*Diff(Pinj)/(4.d0*u0)) !Limits in x (par1)
     Xmax=dlog(X0)  
     Xstep=(Xmax-Xmin)/dfloat(N-1)     
     do i=1,N      
        X(i)=Xmin+dfloat(i-1)*Xstep
     enddo

!!!!	Grid in log(p)              
     Ymin=dlog(Pinj)
     Pstar=X0*u0/DB0	
     Ymax=dlog(Pstar*par2)
     Ystep=(Ymax-Ymin)/dfloat(N-1)
     do j=1,N		
        lnp(j)=Ymin+dfloat(j-1)*Ystep
     enddo
     p=dexp(lnp)

!!!!	Diffusion coefficient dvec=Diff(p)*DB0                    
     dummy=max(dsqrt(8.d0*pi*mp*n0*PB(1))*u0,B0)/B0 !Guess dB(1)/B0
     do i=1,N    		
        do j=1,N
           dvec(i,j)=Diff(p(j))*DB0/dummy
           !dvec(i,j)=min(Diff(p(j))*DB0/dummy,1.d28*Diff(p(j))**(1.d0/3.d0))
        enddo
     enddo
     ! write(*,*) dvec
     dvecold=dvec

!!!!	fxp(x,p) & F(p) in the test-particle case - guess
     do j=1,N
        tmp(j)=max(dexp(-u0*x0/dvec(1,j)),1.d-100)
        xvec1(j)=lnp(j)
        yvec1(j)=1.d0/(1.d0-tmp(j))
     enddo
     dummy=eta*n0/pi/Pinj**3.d0
     do j=1,N
        call integrate(2,lnp(1),lnp(j),tmp2(j)) !PROBLEM
        f0old(j)=max(dummy*dexp(-4.d0*tmp2(j)),1.d-100)
        Ftldold(j)=max(tmp(j)*yvec1(j),1.d-100)
     enddo

     do i=1,N
        do j=1,N
           Kxold(i,j)=max((dexp(-u0*dexp(X(i))/dvec(1,j))-tmp(j))*yvec1(j),1.d-100) !f(x,p)/f0(p)
        enddo
     enddo
     ! if (verb.eq.1) then
        ! write(*,*)'ok fxpTP(x,p)'
        ! open(1,file='fxpTP.dat')
        ! do i=1,N
           ! do j=1,N
              ! pp=4.d0*pi*p(j)**4.d0
              ! write(1,*) X(i), p(j),f0old(j)*pp/n0*Kxold(i,j),Kxold(i,j)
           ! enddo
        ! enddo
        ! close(1)
     ! end if

!!!!	csi_c(x) guess
     do i=1,N
        do j=1,N
           xvec(j)=lnp(j)
           yvec(j)=dlog(f0old(j))+dlog(Kxold(i,j))
        enddo
        Y1=Ymin
        do j=1,N
           if(funcxi(lnp(N+1-j)).ge.1.d-100)then
              Y1=lnp(N+1-j)
           endif
        enddo
        call integrate(4,Y1,Ymax,csi_c(i))
     enddo
     
!!!!	Renormalization
     csic0=1.d0-rsrt-PB(1)+(1.d0/(gammag*M0*M0))*(1.d0-(1.d0+Fth)*rsrt**(-gammag))
     res=csic0/csi_c(1)
     csi_c=max(csi_c*res, 1.d-10)
     f0=f0*res

     ! if (verb.eq.1) then
        ! open(1,file='csiTP.dat',access='append') !DEBUG
        ! do j=1,N
           ! write(1,*) X(j),csi_c(j)
        ! enddo
        ! close(1)
     ! end if

 !!!!	PB(x) guess
     dummy=8.d0*pi*mp*n0
     ! PB(1)=csi_c(1)/chi
     ! PB(1)=delta*(1.d0-rsrt)*3.d0/(1.d0+2.d0*Rtot**2.d0) ! hack RD
     do i=1,N
        PB(i)=PB(1) !*csi_c(i)/csi_c(1) !#MFA
        dB(i)=max(dsqrt(dummy*PB(i))*u0,B0)  !#MFA
        phi(i)=dB(i)/B0
        V(i)=dsqrt(1.d0-csi_c(i))*phi(i)/Ma0
     enddo

!!!!	U(x) guess 
     dummy=1.d0/gammag/M0/M0
     do i=1,N
        U(i)=1.d0+dummy*(1.d0-rsrt**(-gammag))-csi_c(i)-PB(i)
     enddo
     !		do i=1,N
     !			write(*,*), U(i), V(i), csi_c(i)
     !		enddo

!!!!!!!!!!! Scattering center compression ratio
	 
 !!!!	hack
	 !V2 = dsqrt(delta*2.d0*csic0/Rtot) ! hack s.t. PB2 = Pcr*delta
	 V2=dsqrt(2.d0*PB(1)/3.d0/Rtot*(1.d0+2.d0*Rsub**2.d0))
	 
     !Ssub=(U(1)-V(1))/(U(1)/Rsub)
     !Stot=(U(N)-V(N))/(U(1)/Rsub)
     !Ssub = (U(1)-V(1))/(U(1)/Rsub+V2)
     !Stot = (U(N)-V(N))/(U(1)/Rsub+V2)
      Ssub = (U(1))/(U(1)/Rsub+V2)
      Stot = (U(N))/(U(1)/Rsub+V2)
     
!!!!	Up(p) guess
     fi = (U-V)
     call grad(fi,X,DU)
     do i=1,N
        DU(i)=max(DU(i),1.d-100)
     enddo
     do j=1,N
        do i=1,N
           xvec1(i)=X(i)
           yvec1(i)=DU(i)*Kxold(i,j)
        enddo
        call integrate(2,X(1),X(N),tmp(j))
        Upold(j)=U(1)-V(1)+tmp(j)
     enddo
     ! if (verb.eq.1) then
        ! write(*,*)'ok UpTP'
        ! open(1,file='UpTP.dat',access='append') !DEBUG
        ! do j=1,N
           ! write(1,*) p(j),Upold(j),U(j),V(j), DU(j)
        ! enddo
        ! close(1)
     ! end if

     write(*,*) 'rsrt=',rsrt, 'Rtot=',Rtot 

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!    
     !     Iteration at fixed Rsub/Rtot
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!         
     rd0=1.d-2 ! #CONTROL Convergence
     Nprove=0  
     jf=0
     jn=0

     do while((Nprove.le.100.and.jf.le.4)) 
        Nprove=Nprove+1
        !	       	write(*,*) Nprove
        csic0=1.d0-rsrt+(1.d0/(gammag*M0*M0))*(1.d0-(1.d0+Fth)*rsrt**(-gammag))-PB(1)	
        !			write(*,*) 'xi0=',csic0,'err=',-(csic0-csic0_err)/csic0

!!!! DU/Dx
        fi = (U-V)
        call grad(fi,X,DU)
        do i=1,N
           DU(i)=max(DU(i),1.d-100)
        enddo
        ! if (verb.eq.1) then
           ! write(*,*)'ok DU'
           ! if(Nprove.eq.1)then
              ! open(1,file='du.dat',access='append')
              ! do i=1,N
                 ! write(1,*) X(i),DU(i), U(i), V(i)
              ! enddo
              ! close (1)
           ! endif
        ! end if

!!!!	Up(p)
        do j=1,N
           do i=1,N
              xvec1(i)=X(i)
              yvec1(i)=DU(i)*Kxold(i,j)
           enddo
           call integrate(2,X(1),X(N-1),tmp(j))
           Up(j)=U(1)-V(1)+tmp(j)
        enddo
        ! if (verb.eq.1) then
           ! write(*,*)'ok Up(p)'
           ! open(1,file='Up.dat',access='append')
           ! do j=1,N
              ! write(1,*) p(j),Up(j), Upold(j)
           ! enddo
           ! close(1)
        ! endif

!!!!	Psi(x,p)
        do j=1,N
           do i=1,N
              xvec(i)=X(i)
              yvec(i)=dlog(U(i)-V(i))-dlog(dvec(i,j))+dlog(u0)
            enddo
           do i=1,N
              call integrate(1,Xmin,X(i),exppsi) !#CONTROL tol*1e3
              Psi(i,j)=-min(exppsi,2.d2)
           enddo
        enddo
        ! if (verb.eq.1) then
           ! write(*,*)'ok Psi(x,p)'
           ! if(Nprove.eq.1)then
              ! open(1,file='psiD.dat',access='append')
              ! do i=1,N
                 ! do j=1,N
                    ! write(1,*) X(i), p(j), psi(i,j), dvec(i,j)
                 ! enddo
              ! enddo
              ! close(1)
           ! endif
        ! endif

!!!!	f0(p)
        dummy=eta*N0/4.d0/pi/Pinj**3.d0
        do j=1,N
           xvec1(j)=lnp(j) 
           yvec1(j)=3.d0*Stot/(Upold(j)*Stot-1.d0)*(Upold(j)+Ftldold(j))
        enddo
        do j=1,N
           call integrate(2,lnp(1),lnp(j),exppsi)
           f0(j)=max(dummy*3.d0*Stot/(Stot*Upold(j)-1.d0)*dexp(-exppsi),1.d-100) !#CONTROL
        enddo

        ! if (verb.eq.1) then
           ! write(*,*)'ok f0(p)'
        ! end if
        !open(1,file='f0.dat')!,ACCESS='append'
        !do j=1,N
        !	pp=p(j)**4.d0
        !	write(1,*) p(j), f0(j)*pp!, Up(j),Upold(j)
        !enddo
        !close(1)

!!!!	Wxp(x,p)
        do j=1,N
           do i=1,N
              xvec(i)=X(i)
              yvec(i)=-Psi(i,j)-dlog(dvec(i,j))
           enddo
           do i=1,N ! B(x,y) is the W(x,p) in the paper
              call integrate(1,X(1),X(i),Bxp(i,j)) !#CONTROL #SPEED
           enddo
           do i=1,N
              Wxp(i,j)=max(Bxp(i,j)/Bxp(N,j),1.d-100)  !This is W(x,p)/W0(p)
           enddo
        enddo
        ! if (verb.eq.1) then
           ! write(*,*)'ok Wxp(x,p)'
        ! end if

!!!!	F(p) 
        do j=1,N
           Ftld(j)=max(1.d0/Bxp(N,j)/u0,1.d-100)
        enddo
        ! if (verb.eq.1) then
           ! write(*,*)'ok F(p)'
           ! if(Nprove.eq.1)then
              ! open(1,file='Fp.dat',access='append')
              ! do j=1,N
                 ! pp=4.d0*pi*p(j)**4.d0
                 ! write(1,*) p(j),Ftldold(j),Ftld(j)!*f0(j)*pp/n0
              ! enddo
              ! close(1)
           ! end if
        ! endif
     


!!!!	f(x,p)
        do i=1,N
           do j=1,N
              Kx(i,j)=max(dexp(Psi(i,j))*(1.d0-Wxp(i,j)),1.d-100)
              fxp(i,j)=f0old(j)*Kx(i,j)
           enddo
        enddo
        ! if (verb.eq.1) then
           ! write(*,*)'ok f(x,p)'
           ! if(Nprove.eq.1)then
              ! open(1,file='fxp.dat',access='append')
              ! do i=1,N
                 ! do j=1,N
                    ! pp=p(j)**4.d0
                    ! write(1,*) X(i), p(j),fxp(i,j)*pp,Kx(i,j)
                 ! enddo
              ! enddo
              ! close(1)
           ! endif
        ! endif

!!!!	csi_c(x)
        dummy=4.d0*pi/3.d0/n0/mp/u0/u0*mc*c
        do i=1,N-1
           do j=1,N
              xvec(j)=lnp(j)
              yvec(j)=dlog(fxp(i,j)) 
           enddo
           Y1=Ymin
           do j=1,N
              if(funcxi(lnp(N+1-j)).ge.1.d-100)then 
                 Y1=lnp(N+1-j)
              endif
           enddo
           if(Y1.ge.Ymax)then
             csi_c(i)=1.d-100
           else
              call integrate(4,Y1,Ymax,csi_c(i))
           end if  
           csi_c(i)=csi_c(i)*dummy
        enddo
        csi_c(N)=1.d-100
        csic0_err=csi_c(1)
        
        ! if (verb.eq.1) then
           ! write(*,*)'ok csi_c(x)'
           ! open(1,file='csi.dat',access='append') !DEBUG
           ! do i=1,N
              ! write(1,*) X(i),csi_c(i)
           ! enddo
           ! close(1)
        ! endif
        ! write(*,*) 'xi(1)=',csi_c(1)

!!!! Fesc
        do j=1,N
           xvec(j)=lnp(j)
           yvec(j)=dlog(Ftld(j))+dlog(f0(j))+dlog(u0)
        enddo
        Y1=Ymin
        do j=1,N
           if(funcxi(lnp(N+1-j)).gt.1.d-100)then 
              Y1=lnp(N+1-j)
           endif
        enddo
        call integrate(4,Y1,lnp(N),Fesc)
        
        Fesc=Fesc/mp/n0/u0/u0/u0*2.d0*4.d0*pi*mc*c
        ! write(*,*) 'Fesc=', Fesc

!!!!	New PB and Alfvèn velocity
        dummy= Fesc/c/2.d0*u0 ! #MFA: Bell's instability
        !write(*,*)  'Fesc=', Fesc,  'PB0=', dummy
        do i=1,N
           !PB(i)=2.d0*(1.d0-U(1)**(5.d0/4.d0))**2.d0/25.d0/U(1)**1.5d0*(1.d0-zeta)*delta ! #MFA
           !PB(i)=u0/2.d0/c*(1.d0-rsrt)/(gammaCR-1.d0)*delta
           PB(i)=3.d0/2.d0*(1.d0-rsrt)/Ma0*delta
           ! PB(i)=3.d0/(1.d0+2.d0*Rsub**(2.d0))*csic0 ! PB2 = csi_c(1); upstream is decompressed
           !PB(i)=delta*csic0*3.d0/(1.d0+2.d0*Rsub**2.d0)! hack to remove upstream contribution (37)
           dB(i)=max(dsqrt(8.d0*pi*mp*n0*PB(i))*u0,B0)
           phi(i)=dB(i)/B0 
           V(i)=dsqrt(U(i))*phi(i)/Ma0            ! #MFA
        enddo
!!$                do i=1, N
!!$                	write(*,*) U(i), V(i), PB(i), phi(i)
!!$                enddo 

!!!!	New Ssub/Stot
 
 !!!!	hack
	 	!V2 = dsqrt(delta*2.d0*csic0/Rtot) ! hack s.t. PB2 = Pcr*delta
	 	V2=dsqrt(2.d0*PB(1)/3.d0/Rtot*(1.d0+2.d0*Rsub**2.d0))
	 
     	!Ssub=(U(1)-V(1))/(U(1)/Rsub)
     	!Stot=(U(N)-V(N))/(U(1)/Rsub)
     	!Ssub = (U(1)-V(1))/(U(1)/Rsub+V2)
     	!Stot = (U(N)-V(N))/(U(1)/Rsub+V2)
     	Ssub = (U(1))/(U(1)/Rsub+V2)
     	Stot = (U(N))/(U(1)/Rsub+V2)
     	! write(*,*) 'Ssub=',Ssub, 'Stot=',Stot


!!!!	New D(x,p)
        do i=1,N    		
           do j=1,N             
              dvec(i,j)=min(Diff(p(j))*DB0/phi(i),1.d28*Diff(p(j))**(1.d0/3.d0))
              dvecold(i,j)=dvec(i,j)
           enddo
        enddo

!!!!	Variable update
        Upold=(Up+Upold)/2.d0
        f0old=(f0+f0old)/2.d0
        Ftldold=(Ftld+Ftldold)/2.d0
        do i=1,N
           do j=1,N
              Kxold(i,j)=(Kx(i,j)+Kxold(i,j))/2.d0
           enddo
        enddo

!!!!	Next step of the iteration for solving the system of equations              
        restold=rest
        rest=(csi_c(1)-csic0)/csic0 
        rdiff=dabs(rest-restold) 
        if(rdiff.le.rd0)then 	
           if(jf.eq.0)then  
              jf=1
              jn=Nprove
           else if(jn.eq.Nprove-1)then
              jf=jf+1
              jn=Nprove
           else
              jf=0
           endif
        endif

!!!!	Convergence renormalization
        res=csic0/csi_c(1)
        csi_c=csi_c*res
        f0=f0*res

!!!!	New tau(x), U(x) 
        do i=1,N
           tau(i)=1.d0/gammag/M0/M0/U(i)**gammag &!&
                *(1.d0+zeta*(gammag-1.d0)*M0*M0/Ma0*(1.d0-U(i)**gammag)) !#PHYSICS : Ma0, not Ma(X)
           U(i)=1.d0+tau(N)-csi_c(i)-tau(i)-PB(i)
        enddo
        !open(1,file='ux.dat')!,ACCESS='append')
        !do i=1,N
        !	write(1,*) dexp(X(i)),U(i)
        !enddo
        !close(1)
        !	write(*,*) 'Nprove=',Nprove, 'jf=',jf	           
     enddo ! End of the itration on Nprove		

!!!!     New iteration on Rsub/Rtot
     if(rest.gt.0.d0)then
        rsrt2=rsrt
        rest2=rest
     else
        rsrt1=rsrt
        rest1=rest
     endif
     if(rest1.ne.0.d0 .and. rest2.ne.0.d0)then
        rsrt=rsrt1-rest1*(rsrt2-rsrt1)/(rest2-rest1)
     else
        rsrt=(rsrt2+rsrt1)/2.d0
     endif
     if(dabs(rsrt-rsrtold)/rsrt.le.1.d-6) then
        !write(*,*)  'ERROR: Try a different choice of RsRt_min'
        stop  
     else
        rsrtold=rsrt
        arest=dabs(rest)
     endif

  enddo ! End of the iteration on arest (on Rs/Rt)

!!!!! FINAL OUTPUT
  do j=1,N
     pp=4.d0*pi*p(j)**4.d0
     p_outMatrix(timestep,j)=p(j)
     f0_outMatrix(timestep,j)=f0(j)*pp/n0
     fesc_outMatrix(timestep,j)=Ftld(j)*f0(j)*pp/n0 !this is not dimensionless
  enddo
  !open(1,file='hydro.dat',ACCESS='append')
  ! do i=1,N
     ! write(1,*) X(i), U(i), csi_c(i), PB(i), tau(i)
  ! enddo
  !close(1)


!!!!	Downstream magnetic field
  B2=dB(1)*Rsub
  Pcr=csi_c(1)
  
  
  
  !dumping into ararys
  PcrOutputArray(timestep) = Pcr
  B2OutputArray(timestep) = B2
  T2OutputArray(timestep) = T2
  RtotOutputArray(timestep) = Rtot
  FescOutputArray(timestep) = Fesc

!!!!	Maxwell-Boltzmann distribution for the protons: N points below pth
  !	  pth = p(1)/csi
  !	  pMB(N) = p(1)*mc
  !	  fMB(N) = f0(1)
  !	  pMB(1)   = pMB(N)/(50.d0*csi)
  !	  norm=fMB(N)*dexp(csi*csi)
  !	  fMB(1)   = norm*dexp(-(1.d0/50.d0)**2)
  !
  !	  DO j = 2, N-1
  !		pMB(j) = pMB(1) *(pMB(N)/pMB(1))**(dfloat(j-1)/dfloat(N-1))
  !		fMB(j) = norm*dexp(-(pMB(j)/pth)**2)
  !	  ENDDO

!!!!		Final output
  !  open(37,FILE='ST1.dat',ACCESS='append')
  !  write(37,'(f8.5,2x,f12.2,2x,f16.6,2x,f6.4,2x,f7.4,2x,f6.4,2x,f7.4,2x,f8.5,2x,f10.6)') &!&
  !       u0/1.d8,Pstar,B2*1.d6, &!&
  !       Rsub,Rtot,Ssub,Stot,csi_c(1),Fesc
  !  close(37)

  return
end subroutine DSA_X0

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!	Function and subroutines for DSA_X0 subroutine	                       !!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
attributes(device) function Diff(mom)
  implicit none
  real*8 :: Diff,mom
  Diff=(mom*mom/dsqrt(1.d0+mom*mom))
  return
end function Diff

attributes(device) function funcexp(z)
  implicit none
  integer N
  parameter(N=50)
  real*8 :: funcexp,z,divdif,xvec(N),yvec(N)
  COMMON/VEC/xvec,yvec
  funcexp=dexp(divdif(yvec,xvec,N,z,3))*dexp(z) 
  return
end function funcexp

attributes(device) real*8 function funcp1(z)
  implicit none
  integer N
  parameter(N=50)
  real*8 :: funcp1,xvec1(N),yvec1(N),divdif,z
  COMMON/VEC1/xvec1,yvec1
  funcp1=divdif(yvec1,xvec1,N,z,3)
  return
end function funcp1

attributes(device) real*8 function funcp0(z)
  implicit none
  integer N
  parameter(N=50)
  real*8 :: funcp0,xvec0(N),yvec0(N),divdif,z
  COMMON/VEC0/xvec0,yvec0
  funcp0=dexp(divdif(yvec0,xvec0,N,z,3))
  return
end function funcp0

attributes(device) real*8 function funcxi(z) 
  implicit none
  integer N
  parameter(N=50)
  real*8 :: funcxi,z,xvec(N),yvec(N),divdif
  COMMON/VEC/xvec,yvec
  funcxi=dexp(divdif(yvec,xvec,N,z,3))*dexp(4.d0*z)*dexp(z)/dsqrt(1.d0+dexp(2.d0*z))
  return
end function funcxi

attributes(device) subroutine grad(fi,xi,dfi)
  implicit none
  integer N,i
  parameter(N=50)
  real*8 :: fi(N),xi(N),dfi(N)

  do i=2,N-1
     dfi(i)=(fi(i+1)-fi(i-1))/(xi(i+1)-xi(i-1))
  enddo
  dfi(1)=(fi(2)-fi(1))/(xi(2)-xi(1))
  dfi(N)=(fi(N)-fi(N-1))/(xi(N)-xi(N-1))
  return
end subroutine grad


!!!!	Interpolating subroutine
attributes(device) function DIVDIF(F,A,NN,X,MM)
  implicit none
  integer NN,N,MM,M,MPLUS,IX,IY,MID,NPTS,IP,L,ISUB,J,I,MMAX, errorNumber
  real*8 :: SUM,DIVDIF
  real*8 :: A(NN),F(NN),T(20),D(20),X
  LOGICAL EXTRA

  DATA MMAX/10/
  !     
  !     TABULAR INTERPOLATION USING SYMMETRICALLY PLACED ARGUMENT POINTS.
  !     
  !     START.  FIND SUBSCRIPT IX OF X IN ARRAY A.
  IF( (NN.LT.2) .OR. (MM.LT.1) ) GO TO 20
  N=NN
  M=MM
  MPLUS=M+1
  IX=0
  IY=N+1
  !     (SEARCH INCREASING ARGUMENTS.)
1 MID=(IX+IY)/2
  IF(X.GE.A(MID)) GO TO 2
  IY=MID
  GO TO 3
  !     (IF TRUE.)
2 IX=MID
3 IF(IY-IX.GT.1) GO TO 1
  GO TO 7
  !     COPY REORDERED INTERPOLATION POINTS INTO (T(I),D(I)), SETTING
  !     *EXTRA* TO TRUE IF M+2 POINTS TO BE USED.
7 NPTS=M+2-MOD(M,2)
  IP=0
  L=0
  GO TO 9
8 L=-L
  IF(L.GE.0) L=L+1
9 ISUB=IX+L
  IF((1.LE.ISUB).AND.(ISUB.LE.N)) GO TO 10
  !     (SKIP POINT.)
  NPTS=MPLUS
  GO TO 11
  !     (INSERT POINT.)
10 IP=IP+1
  T(IP)=A(ISUB)
  D(IP)=F(ISUB)
11 IF(IP.LT.NPTS) GO TO 8
  EXTRA=NPTS.NE.MPLUS
  !     
  !     REPLACE D BY THE LEADING DIAGONAL OF A DIVIDED-DIFFERENCE TABLE, SUP-
  !     PLEMENTED BY AN EXTRA LINE IF *EXTRA* IS TRUE.
  DO 14 L=1,M
     IF(.NOT.EXTRA) GO TO 12
     ISUB=MPLUS-L
     D(M+2)=(D(M+2)-D(M))/(T(M+2)-T(ISUB))
12   I=MPLUS
     DO 13 J=L,M
        ISUB=I-L
        D(I)=(D(I)-D(I-1))/(T(I)-T(ISUB))
        I=I-1
13      CONTINUE
14      CONTINUE
        !     
        !     EVALUATE THE NEWTON INTERPOLATION FORMULA AT X, AVERAGING TWO VALUES
        !     OF LAST DIFFERENCE IF *EXTRA* IS TRUE.
        SUM=D(MPLUS)
        IF(EXTRA) SUM=0.5*(SUM+D(M+2))
        J=M
        DO 15 L=1,M
           SUM=D(J)+(X-T(J))*SUM
           J=J-1
15         CONTINUE

           DIVDIF=SUM

           RETURN

20         errorNumber = 1
           errorNumber = 2

           DIVDIF=999999999999999999.

101        errorNumber = 3
102        errorNumber = 4

        END

        subroutine get_time(time_sec) 
          implicit integer(i-n), double precision(a-h, o-z)
          character (len=12) tdum1, tdum2, tdum3
          dimension i_date(8)
          call date_and_time(tdum1, tdum2, tdum3, i_date)
          time_sec = i_date(3)*24.0*3600.0 + i_date(5)*3600.0
          time_sec = time_sec + i_date(6)*60.0 + i_date(7)+ i_date(8)/1.d3
          return
        END subroutine get_time

!!!!------------------------------------------------------------------------------
!!!!		Subroutine for integrals 
        attributes(device) SUBROUTINE integrate(func,a,b,ss) 
          INTEGER KMAX,KMAXP,J,JM
          real*8 :: a,b,ss,tol
          integer :: func
          PARAMETER (KMAX=14, KMAXP=KMAX+1, J=5, JM=J-1, tol=1.d-1) !#CONTROL Should work even for EPS=1e-1
          INTEGER k
          real*8 :: dss,h(KMAXP),s(KMAXP)

          h(1)=1.d0
          do k=1,KMAX 
             call halfway(func,a,b,s(k),k)
             if (k.ge.J) then
                call pol_interpol(h(k-JM),s(k-JM),J,0.d0,ss,dss)
                if (dabs(dss).le.tol*dabs(ss)) return
             endif
             s(k+1)=s(k)
             h(k+1)=h(k)/9.d0
          end do
          !write(*,*)  'ERROR in subroutine Integrate'
          stop
        END SUBROUTINE integrate

        attributes(device) SUBROUTINE pol_interpol(xa,ya,n,x,y,dy)
          INTEGER n,NMAX
          real*8 :: dy,x,y,xa(n),ya(n)
          PARAMETER (NMAX=10)
          INTEGER i,m,ns
          real*8 :: den,dif,dift,ho,hp,w,c(NMAX),d(NMAX)
          ns=1
          dif=dabs(x-xa(1))
          do i=1,n
             dift=abs(x-xa(i))
             if (dift.lt.dif) then
                ns=i
                dif=dift
             endif
             c(i)=ya(i)
             d(i)=ya(i)
          end do
          y=ya(ns)
          ns=ns-1
          do m=1,n-1
             do i=1,n-m
                ho=xa(i)-x
                hp=xa(i+m)-x
                w=c(i+1)-d(i)
                den=ho-hp
                if(den.eq.0.d0)then
                   !write(*,*)  'ERROR in subroutine interpol'
                   stop
                end if
                den=w/den
                d(i)=hp*den
                c(i)=ho*den
             end do
             if (2*ns.lt.n-m)then
                dy=c(ns+1)
             else
                dy=d(ns)
                ns=ns-1
             endif
             y=y+dy
          end do
          return
        END SUBROUTINE pol_interpol

      attributes(device) SUBROUTINE halfway(func,a,b,s,n)
      INTEGER n
      real*8 :: a,b,s
      integer :: func
      INTEGER :: it,j
      real*8 :: ddel,del,sum,tnm,x
      
      Select Case (func)
      
        case (1)
          if (n.eq.1) then
            s=(b-a)*funcexp(0.5d0*(a+b))
          else
            it=3**(n-2)
            tnm=it
            del=(b-a)/(3.d0*tnm)
            ddel=del+del
            x=a+0.5d0*del
            sum=0.d0
            do j=1,it
              sum=sum+funcexp(x)
              x=x+ddel
              sum=sum+funcexp(x)
              x=x+del
           end do
            s=(s+(b-a)*sum/tnm)/3.d0
          endif
        case (2)
          if (n.eq.1) then
            s=(b-a)*funcp1(0.5d0*(a+b))
          else
            it=3**(n-2)
            tnm=it
            del=(b-a)/(3.d0*tnm)
            ddel=del+del
            x=a+0.5d0*del
            sum=0.d0
            do j=1,it
              sum=sum+funcp1(x)
              x=x+ddel
              sum=sum+funcp1(x)
              x=x+del
           end do
            s=(s+(b-a)*sum/tnm)/3.d0
          endif
        case (3)
          if (n.eq.1) then
            s=(b-a)*funcp0(0.5d0*(a+b))
          else
            it=3**(n-2)
            tnm=it
            del=(b-a)/(3.d0*tnm)
            ddel=del+del
            x=a+0.5d0*del
            sum=0.d0
            do j=1,it
              sum=sum+funcp0(x)
              x=x+ddel
              sum=sum+funcp0(x)
              x=x+del
           end do
            s=(s+(b-a)*sum/tnm)/3.d0
          endif
        case (4)
          if (n.eq.1) then
            s=(b-a)*funcxi(0.5d0*(a+b))
          else
            it=3**(n-2)
            tnm=it
            del=(b-a)/(3.d0*tnm)
            ddel=del+del
            x=a+0.5d0*del
            sum=0.d0
            do j=1,it
              sum=sum+funcxi(x)
              x=x+ddel
              sum=sum+funcxi(x)
              x=x+del
           end do
            s=(s+(b-a)*sum/tnm)/3.d0
          endif
        case default
            return
      end select
      return
      END SUBROUTINE halfway

end module GPUFunctions








program CRAFT_RSN

  ! Modified version of CRAFT in which PB(1) = 1.5*P_CR/M_A 
  ! (see Caprioli and Spitkovsky 2014 II)
  ! fudge factor (delta) on PB is included (v2)
  use cudafor
  use GPUFunctions
  
  implicit none 

  integer N,j, numberOfTimesteps, threadsN
  PARAMETER (N=50)

  real*8 :: M0,B0,T0,N0,xi_inj,X0, Pcr_guess, Pcr_max, delta
  real*8 :: Pcr,T2,Fesc,B2,Rtot
  real*8 :: age, Rsh, Vsh
  real*8 :: time_end_sec,time_begin_sec,run_time_sec
  real*8 :: mp,c,kB,qe,pc,pi
  !Evo parameters
  real*8 :: E51,Mej,age1,age2
  real*8 :: TS1,TS2,agestep,TSedov,ageS,Nage
  
  
  
  character text*10
  
  ! real*8, device :: M0,B0,T0,N0,xi_inj,X0, Pcr_guess, Pcr_max, delta
  ! real*8, device ::Pcr,T2,Fesc,B2,Rtot
  ! real*8, device ::age, Rsh, Vsh
  ! real*8, device ::time_end_sec,time_begin_sec,run_time_sec
  ! real*8, device ::mp,c,kB,qe,pc,pi
  ! !Evo parameters
  ! real*8, device ::E51,Mej,age1,age2
  ! real*8, device ::TS1,TS2,agestep,TSedov,ageS,Nage
  
  !these are the arrays that need to be allocated by file size and N
  real*8, dimension (:), allocatable :: ageArray, RshArray, VshArray, n0Array, B0Array !initial creation and size of input arrays
  real*8, dimension (:), allocatable, device :: ageArray_d, RshArray_d, VshArray_d, n0Array_d, B0Array_d !initial creation of device arrays
  
  real*8, dimension (:), allocatable :: PcrOutputArray, FescOutputArray, RtotOutputArray, T2OutputArray, B2OutputArray !output arrays for passing up and down
  real*8, dimension (:), allocatable, device :: PcrOutputArray_d, FescOutputArray_d, RtotOutputArray_d, T2OutputArray_d, B2OutputArray_d !device version REMEMBER TO ALIGN TIMESTEP WITH INDEX
  
  real*8, dimension (:,:), allocatable :: p_outMatrix, f0_outMatrix, fesc_outMatrix !Probably gonna be timestep * N size. Just need to remember how to correctly read through them when I writeout
  real*8, dimension (:,:), allocatable, device :: p_outMatrix_d, f0_outMatrix_d, fesc_outMatrix_d !I deadass don't know how we're going to have to index this. Probably blockId for timestep and then threadId<=N for N loop? What a nightmare

  PARAMETER (mp=1.67d-24,c=3.d10,kB=1.38d-16,qe=4.8d-10,pc=3.1d18,pi=3.1415d0)


  !allocation of the arrays
  do
    read (3,*,end = 12) age
    numberOfTimesteps = numberOfTimesteps+1
  enddo 
  12 close(3)
  print *, numberOfTimesteps
  
  !these are the inputs
  allocate(ageArray(numberOfTimesteps))
  allocate(RshArray(numberOfTimesteps))
  allocate(VshArray(numberOfTimesteps))
  allocate(N0Array(numberOfTimesteps))
  allocate(B0Array(numberOfTimesteps))
  allocate(ageArray_d(numberOfTimesteps))
  allocate(RshArray_d(numberOfTimesteps))
  allocate(VshArray_d(numberOfTimesteps))
  allocate(n0Array_d(numberOfTimesteps))
  allocate(B0Array_d(numberOfTimesteps))
  
  !these are the array outputs
  !                                            PcrOutputArray, T2OutputArray, B2OutputArray, FescOutputArray, RtotOutputArray
  
  allocate(PcrOutputArray(numberOfTimesteps))
  allocate(FescOutputArray(numberOfTimesteps))
  allocate(RtotOutputArray(numberOfTimesteps))
  allocate(T2OutputArray(numberOfTimesteps))
  allocate(B2OutputArray(numberOfTimesteps))
  allocate(PcrOutputArray_d(numberOfTimesteps))
  allocate(FescOutputArray_d(numberOfTimesteps))
  allocate(RtotOutputArray_d(numberOfTimesteps))
  allocate(T2OutputArray_d(numberOfTimesteps))
  allocate(B2OutputArray_d(numberOfTimesteps))
  
  
  
  !these are the matrix outputs
  allocate(p_outMatrix(numberOfTimesteps,N))
  allocate(f0_outMatrix(numberOfTimesteps,N))
  allocate(fesc_outMatrix(numberOfTimesteps,N))
  allocate(p_outMatrix_d(numberOfTimesteps,N))
  allocate(f0_outMatrix_d(numberOfTimesteps,N))
  allocate(fesc_outMatrix_d(numberOfTimesteps,N))
 
  numberOfTimesteps = 1
  open(3, file='Evolution.dat')
  do
    read (3,*,end = 11) age, Rsh, Vsh, n0, B0
    ageArray(numberOfTimesteps) = age
    RshArray(numberOfTimesteps) = Rsh
    VshArray(numberOfTimesteps) = Vsh
    n0Array(numberOfTimesteps) = n0
    B0Array(numberOfTimesteps) = B0
    numberOfTimesteps = numberOfTimesteps+1
  enddo
  11 close(3)


! figure out how to find the closes 32, the usual thing isn't working
!  threadsN = ((N−1)|31)+1 
  threadsN = 64

  call get_time(time_begin_sec) 


!!!! Loading input parameters
  open(2,file='Init.dat')
  read(2,*) text, T0
  !  write(6,'(a,f10.0)') text, T0
  read(2,*) text, TS1,TS2 !obsolete
  read(2,*) text, Nage !obsolete
  read(2,*) text, E51 !obsolete
  read(2,*) text, Mej !obsolete
  read(2,*) text, Pcr_guess, Pcr_max
  read(2,*) text, xi_inj
  read(2,*) text, B0
  read(2,*) text, N0
  read(2,*) text, X0
  read(2,*) text, delta
  close(2)
  
  
  !moving things onto the proper device version 
  ageArray_d = ageArray
  RshArray_d = RshArray
  VshArray_d = VshArray
  N0Array_d = n0Array
  B0Array_d = B0Array
  
  
  


  !here's the moment of truth
  !initial version does calculate the proper number of threads to avoid warp problems, but it only uses one for now

  
  call launchToGPU<<<numberOfTimesteps,threadsN>>>(ageArray_d, RshArray_d, VshArray_d, N0Array_d, B0Array_d, T0, Pcr_guess, Pcr_max, xi_inj, delta, X0, &
                                                   PcrOutputArray_d, T2OutputArray_d, B2OutputArray_d, FescOutputArray_d, RtotOutputArray_d, &
                                                   p_outMatrix_d, f0_outMatrix_d, fesc_outMatrix_d)





  PcrOutputArray = PcrOutputArray_d
  T2OutputArray = T2OutputArray_d
  B2OutputArray = B2OutputArray_d
  FescOutputArray = FescOutputArray_d
  RtotOutputArray = RtotOutputArray_d
  p_outMatrix = p_outMatrix_d
  f0_outMatrix = f0_outMatrix_d
  fesc_outMatrix = fesc_outMatrix_d





                                            ! !inputs arrays, then result output arrays, then result output matrix then values for writing during DSA
  ! attributes(global) subroutine launchToGPU(ageArray, RshArray,VshArray, N0Array, B0Array, T0, Pcr_guess, Pcr_max, xi_inj, delta, X0 &
                                            ! PcrOutputArray, T2OutputArray, B2OutputArray, FescOutputArray, RtotOutputArray, &
                                            ! p_outMatrix, f0_outMatrix, fesc_outMatrix)
    ! implicit none
    ! real*8 :: M0, X0Value !input values
    ! real*8, value :: T0, Pcr_guess, Pcr_max, xi_inj, delta, X0
    ! real*8, dimension(:) :: ageArray, RshArray,VshArray, N0Array, B0Array !input arrays (timestep) to pass as values
    ! real*8, dimension(:) :: PcrOutputArray, T2OutputArray, B2OutputArray, FescOutputArray, RtotOutputArray !result output arrays (timestep) for final output.dat 
    ! real*8, dimension(:,:) :: p_outMatrix, f0_outMatrix, fesc_outMatrix !result output matrix (timestep, N) for spectrum.dat
    ! integer :: n, i, j
    
    
    ! n = blockIdx%x
    ! i threadIdx%x
    ! M0=VshArray(n)*dsqrt(1.67d-24/5.d0*3.d0/1.38d-16/T0)
    ! X0Value = X0*RshArray(n)
    
    ! if(i == 1)then !only launching on first thread for first version.
        ! call DSA_X0(M0,B0Array(n),T0,n0Array(n),xi_inj,X0Value,Pcr_guess,Pcr_max, PcrOutputArray,T2OutputArray,B2OutputArray,FescOutputArray,RtotOutputArray,p_outMatrix,f0_outMatrix,fesc_outMatrix, delta)
    ! endif
 
 
 
 
 ! end subroutine launchToGPU



	
! !!!! SNR Evolution
  ! open(3, file='Evolution.dat')
  ! do  !Cycle on Radius                 
	 ! read (3,*,end = 10) age, Rsh, Vsh, n0, B0
	 
	 ! M0=Vsh*dsqrt(1.67d-24/5.d0*3.d0/1.38d-16/T0)
     ! write(6,'(a,f11.3,a,f8.0,a,f7.3,a,f8.2,a,f6.0)')'Age=',age,'	Vsh=',Vsh/1.d5,'	Rsh=',Rsh/pc,'  M0=',M0	
     ! write(6,'(a,f9.1,a,f10.1,a,f6.2,a,f8.2)')'T0[K]=',T0,' B0[muG]=',B0*1.d6,' X0[pc]=',X0/pc*Rsh
     
! !!!!	Call the subroutine which solves the problem at fixed X0

     ! call DSA_X0(M0,B0,T0,n0,xi_inj,X0*Rsh,Pcr_guess,Pcr_max, Pcr,T2,B2,Fesc,Rtot,p_out,f0_out,fesc_out, delta)

     ! !! Save spectrum in a file	
     ! open(1,file='spectrum.dat',ACCESS='append')
     ! do j=1,N
        ! write(1,*) p_out(j), f0_out(j),fesc_out(j)
     ! enddo

! !!!! Save other quantities in a file
     ! open(37,FILE='output.dat',ACCESS='append')
     ! write(37,'(f9.5,2x,f11.3,2x,f8.5,2x,f16.6,2x,f8.5,2x,f10.6,2x,f9.6,2x,f10.2)') &!&
          ! Rsh/pc,age,Vsh/1.d8,B2*1.d6,Pcr,Fesc,Rtot,n0
     ! close(37)

     ! write(6,'(a,f9.5,a,f8.1,a,f10.6,a,f8.2,/)')'Pcr[%]=',Pcr*1.d2,' B2[muG]=',B2*1.d6,' Fesc[%]=',Fesc*1.d2,' T2[10^6K]=',T2/1.d6		

    ! !Pcr_guess =Pcr ! Guess for next time-step
  ! enddo !End of cycle on the Radius
  ! 10 close(3)

  ! call get_time(time_end_sec)
  ! run_time_sec=time_end_sec-time_begin_sec
  ! write(6,230) run_time_sec, (run_time_sec/60.0)
! 230 format(' Run time(sec)=',1pe10.3, '  minutes=',1pe10.3,/)

end program CRAFT_RSN
